
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               DoGEll (ver. 1.00, 23 September 2014)
%
% This is the main script used for segmenting the fetal head of 90
% ultrasound images from the recent ultrasound grand challenge
% http://www.ibme.ox.ac.uk/challengeus2012/
%
% The method is described in:
% A. Foi, M. Maggioni, A. Pepe, S. Rueda, A. T. Papageorghiou, J. A. Noble,
% and J. Tohka, "Difference of Gaussians revolved along elliptical paths
% for ultrasound fetal head segmentation", to appear in Computerized
% Medical Imaging and Graphics, 2014.
%
% Copyright (c) 2012-2014  Tampere University of Technology All rights
% reserved.
%
% This work should only be used for nonprofit purposes. Any unauthorized
% use of these routines for industrial or profit-oriented activities is
% expressively prohibited. By downloading and/or using any of these files,
% you implicitly agree to all the terms of the TUT limited license, as
% specified in the document LEGAL_NOTICE.txt included in this package.
% Please read the TUT limited license before you proceed with using these
% files.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% Populate variable with ultrasound image(s). They can be in any format
%%% readable by MATLAB imread function, as well as DICOM file.
FilenamesCell=dir('*.png');

ToDo=1:numel(FilenamesCell);   % index of the images in FilenamesCell that we want to process

circumS=zeros(numel(FilenamesCell),1);
CostS=zeros(numel(FilenamesCell),1);
ParamsS=zeros(numel(FilenamesCell),6);
filenameS=cell(numel(FilenamesCell),1);
EvolsS=cell(numel(FilenamesCell),1);
timePreprocessingS=zeros(numel(FilenamesCell),1);
timeOptimizationS=zeros(numel(FilenamesCell),1);

%%% Save results to file
writeImageYellowEllipses=false;
writeImageBinaryEllipse=false;
writeImagePositiveEllipse=false;

PA=100;
PB=100;

for iiii=ToDo
    %%%  load image
    fnamme=FilenamesCell(iiii).name;
    disp(['Processing ',fnamme]);
    try
        z=im2double(rgb2gray(imread(fnamme)));
    catch err
        z=im2double(dicomread(fnamme,'frames',1));
    end
    
    preprocessingTime0=now;
    
    %%% Image Padding
    PadBorderPixel=120;
    z=zPad(z,PadBorderPixel);
    
    %%% diffusion-based extrapolation
    toBeExtrapolated=z~=0;   % region to be extrapolated (assumes zero-values as missing data)
    
    [maskkx, maskky]=meshgrid(linspace(0,1,size(z,2)),linspace(0,1,size(z,1)));
    transitionFrom0to1function=@(x,a,b) (x>a).*(x<=b).*(0.5*(1-cos(pi*((x-a)./(b-a)))))+(x>b);
    
    dMap=size(z,1)*maskky;
    dMap=min(dMap,size(z,2)*maskkx);
    dMap=min(dMap,size(z,2)*(1-maskkx));
    dMap=min(dMap,size(z,1)*(1-maskky));
    
    diffusedWeights=double(toBeExtrapolated);
    diffusedImage=z.*diffusedWeights;
    diffusionCounter=0;
    diffuse=true;
    while diffuse
        dLeft=ceil(max(dMap(diffusedWeights==0)));
        if isempty(dLeft)
            dLeft=0;
        end
        if dLeft>=0
            kernel1DLength=max(3,1+2*round(0.5*(dLeft)/2));
            
            kernel1D=kaiser(kernel1DLength,1);
            kernel1D=kernel1D/sum(kernel1D(:));
            
            diffusedImage=conv2(diffusedImage,kernel1D,'same');
            diffusedImage=conv2(diffusedImage,kernel1D','same');
            diffusedImage(toBeExtrapolated)=z(toBeExtrapolated)*sum(kernel1D)^2;%;/diffusionCounter^2;
            diffusedWeights=conv2(diffusedWeights,kernel1D,'same');
            diffusedWeights=conv2(diffusedWeights,kernel1D','same');
            diffusedWeights(toBeExtrapolated)=1*sum(kernel1D)^2;%/diffusionCounter^2;
            diffusionCounter=diffusionCounter+1;
        end
        if all(diffusedWeights(:)>0)
            diffuse=false;
        end
    end
    z0=diffusedImage./diffusedWeights;
    
    %%% Stabilization of local contrast and range normalization
    transitionPointA=0.002*PA/100*1.13;  % these two constants should be increased to provide stronger correction
    transitionPointB=0.01*PB/100*0.953;
    dctDomainAttenuation=1-transitionFrom0to1function((maskkx+maskky)/2,transitionPointA,transitionPointB);
    zSmooth=idct2(dct2(z0).*dctDomainAttenuation);
    z=z0-zSmooth;
    clear dctDomainAttenuation maskkx maskky zSmooth z0
    z=z-min(z(:));
    z=z/max(z(:));
    z=z*80;
    
    preprocessingTime1=now;
    
    zdct=dct2(z);
    timePreprocessingS(iiii)=(preprocessingTime1-preprocessingTime0)*60*60*24;
    disp(['Preprocessing: ',num2str([iiii timePreprocessingS(iiii)]),' seconds']);
    
    %%% ellipse initialization
    c1=0.5;
    c2=0.5*size(z,2)/size(z,1);
    r1=0.28;
    r2=0.32;
    theta=0;
    thicknessFactor= 0.0186*min((size(z)-2*PadBorderPixel)./size(z));
    EllipseParameters=[c1,c2,r1,r2,theta,thicknessFactor];
    
    %%% downscaling and Nelder-Mead parameters
    downscaleAlpha=sqrt(2); %%%  relative downscaling factor between each downscaling step  (scaling factors are approximately powers of downscaleAlpha)
    minImageSize=28; %%% the smallest downscaled image will have size at least minImageSize x minImageSize;
    downscaleFactor=downscaleAlpha.^fliplr((0:1:floor(log(min(size(z))/minImageSize)/log(downscaleAlpha))));
    downscaleFactor=fliplr(unique(downscaleFactor.*(downscaleFactor>4.0001)+double(downscaleFactor<=4.0001)));
    numberOfNelderMeadRestartsAtEachScale=[500 75 25 1 1 1 ones(1,20) 1];   %%% the last element in the vector is used for the final optimization of the thickness
    MaxNelderMeadIterationsSetEachScale= [45 31 28 25 25*ones(1,20) 7];   %%% the last element in the vector is used for the final optimization of the thickness
    MaxNelderMeadIterationsSetEachScaleAfterFail=MaxNelderMeadIterationsSetEachScale;   %%% the last element in the vector is used for the final optimization of the thickness
    counterLogLine=0;
    
    optimsetdisplay='none';
    fTol=1e-2;
    EllipseParametersCurrentBest=EllipseParameters;
    FunctionValueCurrentBest=10000;
    FVALBase=FunctionValueCurrentBest;
    EvolutionLog=zeros(1,8+5);
    
    counterFunctionEvaluationsForEachScale=zeros(1,numel(downscaleFactor));
    counterIterationsForEachScale=zeros(1,numel(downscaleFactor));
    
    regularizationFunction=@(x,b,m) (x*b+sqrt((2-2*m)^2+(x*b).^2))/(2)+m;
    RandomPermutation=1:5;
    RandomPermutationInverse=1:5;
    
    %%% for each scale of the image
    for scaleIndex=1:numel(downscaleFactor)
        xTol= (2e-03)*downscaleFactor(scaleIndex);
        
        % downscaling
        if downscaleFactor(scaleIndex)~=1
            zRescaled=imresize(z,round(size(z)/downscaleFactor(scaleIndex)));
            disp(['current downscaled image size is  ',num2str(size(zRescaled,2)),' x ',num2str(size(zRescaled,1))])
        else
            zRescaled=z;
            disp(['current image size is  ',num2str(size(zRescaled,2)),' x ',num2str(size(zRescaled,1)),'  (no downscaling)'])
        end
        
        randn('seed',0);
        rand('seed',0);
        
        nelderMeadWasSuccessful=true;
        numberOfNelderMeadRestartsCurrent=numberOfNelderMeadRestartsAtEachScale(scaleIndex);
        
        %%% for each restart of the nelder-mead optimization
        for counterNelderMeadRestarts=1:numberOfNelderMeadRestartsCurrent;
            
            indicatorForParameters=[true true true true true false];  % optimizes only the parameters where the indicator is true
            fitParams=EllipseParameters(indicatorForParameters);
            fixParams=EllipseParameters(~indicatorForParameters);
            
            ParamsScale=10*[ [1 1]*0.1   [1 1]*5   1*3   1 ];
            fitParamsScale=ParamsScale(indicatorForParameters);
            
            % permutate optimization parameters
            RandomPermutation=randperm(nnz(indicatorForParameters));
            RandomPermutationInverse=RandomPermutation;
            for RPcycle=3:factorial(nnz(indicatorForParameters))
                RandomPermutationInverse=RandomPermutationInverse(RandomPermutation);
            end
            
            if counterNelderMeadRestarts==1 && scaleIndex>1
                [FunctionValueCurrentBest]=evalEllipseFit(zRescaled,fitParams,fixParams,indicatorForParameters);
                counterFunctionEvaluationsForEachScale(scaleIndex)=counterFunctionEvaluationsForEachScale(scaleIndex)+1;
                counterLogLine=counterLogLine+1;
                EvolutionLog(counterLogLine,:)=[scaleIndex downscaleFactor(scaleIndex) 0 FunctionValueCurrentBest counterIterationsForEachScale(scaleIndex) counterFunctionEvaluationsForEachScale(scaleIndex) (now-preprocessingTime1)*60*60*24 EllipseParametersCurrentBest];
                disp(num2str(EvolutionLog(counterLogLine,:)));
            end
            
            if nelderMeadWasSuccessful
                MaxNelderMeadIterationsSet=MaxNelderMeadIterationsSetEachScale(scaleIndex);
            else
                MaxNelderMeadIterationsSet=MaxNelderMeadIterationsSetEachScaleAfterFail(scaleIndex);
            end
            
            MaxFunEvalOptSet=2*MaxNelderMeadIterationsSet;  %%% usually this is enough
            
            
            % NELDER-MEAD
            if MaxNelderMeadIterationsSet==0;
                FunctionValue=evalEllipseFit(zRescaled,fitParams,fixParams,indicatorForParameters);
                
                counterIterationsForEachScale(scaleIndex)=counterIterationsForEachScale(scaleIndex)+1;
                counterFunctionEvaluationsForEachScale(scaleIndex)=counterFunctionEvaluationsForEachScale(scaleIndex)+1;
            else
                [fitParams,FunctionValue,OUTPUT]=optimizeEllipseParam(zRescaled,fitParams(RandomPermutation)./fitParamsScale(RandomPermutation),fixParams,fitParamsScale,int8(RandomPermutationInverse - 1), xTol,fTol,MaxNelderMeadIterationsSet,MaxFunEvalOptSet);
                
                fitParams=fitParams(RandomPermutationInverse).*fitParamsScale;
                counterIterationsForEachScale(scaleIndex)=counterIterationsForEachScale(scaleIndex)+OUTPUT; %.iterations;
                counterFunctionEvaluationsForEachScale(scaleIndex)=counterFunctionEvaluationsForEachScale(scaleIndex)+OUTPUT; %.funcCount;
            end
            
            EllipseParameters(indicatorForParameters)=fitParams;
            EllipseParameters(3:4)=abs(EllipseParameters(3:4));  % enforces positivity of the radius parameters
            
            if FunctionValue<FunctionValueCurrentBest  %%% fminsearch succeeded in improving the solution
                FunctionValueCurrentBest=FunctionValue;
                EllipseParametersCurrentBest=EllipseParameters;
                nelderMeadWasSuccessful=true;
            else
                nelderMeadWasSuccessful=false;
            end
            
            
            if nelderMeadWasSuccessful || counterNelderMeadRestarts==1 || numberOfNelderMeadRestartsCurrent==counterNelderMeadRestarts;
                counterLogLine=counterLogLine+1;
                EvolutionLog(counterLogLine,:)=[scaleIndex downscaleFactor(scaleIndex) counterNelderMeadRestarts FunctionValueCurrentBest counterIterationsForEachScale(scaleIndex) counterFunctionEvaluationsForEachScale(scaleIndex) (now-preprocessingTime1)*60*60*24 EllipseParametersCurrentBest];
                disp(num2str(EvolutionLog(counterLogLine,:)));
            end
            
            % perturbations
            if counterNelderMeadRestarts<numberOfNelderMeadRestartsCurrent
                redoPerturbation=true;
                while redoPerturbation
                    EllipseParameters=EllipseParametersCurrentBest;
                    ParamsR=EllipseParameters;
                    P1=max(abs(EllipseParameters(1)-[0 1]))-min(EllipseParameters(3:4));
                    P2=max(abs(EllipseParameters(2)-[0 1]*size(z,2)/size(z,1)))-min(EllipseParameters(3:4));
                    P1r=0.5*P1*(rand.^3*(sign(rand-0.5)));
                    P2r=0.5*P2*(rand.^3*(sign(rand-0.5)));
                    ParamsR(1)=EllipseParameters(1)+P1r;
                    ParamsR(2)=EllipseParameters(2)+P2r;
                    ParamsR(3)=EllipseParameters(3)*regularizationFunction(randn.^1,.5,(0.015+0.75*EllipseParameters(6)*10)/EllipseParameters(3));
                    ParamsR(4)=EllipseParameters(4)*regularizationFunction(randn.^1,.5,(0.015+0.75*EllipseParameters(6)*10)/EllipseParameters(3));
                    ParamsR(5)=EllipseParameters(5)+(2*pi*randn*0.02);
                    
                    % check whether perturbation is admissible! (if not admissible, it will try again until it finds an admissible  pertubation)
                    if min(abs(ParamsR(3:4)))<EllipseParameters(6)*9;   % check that ellipse does not collapse onto itself
                        redoPerturbation=true;
                    else
                        if min([ParamsR(1) 1-ParamsR(1) [ParamsR(2) 1-ParamsR(2)]*size(z,2)/size(z,1)])<(min(ParamsR(3:4))+EllipseParameters(6)*8)  %%% check that ellipse is inside the image
                            redoPerturbation=true;
                        else
                            if max(ParamsR(3:4))>0.7 || min(ParamsR(3:4))>0.5
                                redoPerturbation=true;
                            else
                                redoPerturbation=false;
                            end
                        end
                    end
                end
                if nelderMeadWasSuccessful
                    EllipseParameters=ParamsR;
                else
                    EllipseParameters=ParamsR;
                end
            end
        end
        EllipseParameters=EllipseParametersCurrentBest;
    end
    EllipseParameters=EllipseParametersCurrentBest;
    
    %%% adjust thickness
    disp('adjusting thickness')
    indicatorForParameters=[false false false false false true];
    MaxNelderMeadIterationsSet=MaxNelderMeadIterationsSetEachScale(end);
    MaxFunEvalOptSet=2*MaxNelderMeadIterationsSet;
    xTol=5e-03;
    for jijij=1:numberOfNelderMeadRestartsAtEachScale(end)
        fTol=5e-3;
        fitParams=EllipseParameters(indicatorForParameters);
        fixParams=EllipseParameters(~indicatorForParameters);
        [fitParams,FunctionValueCurrentBest,EXITFLAG,OUTPUT]=NelderMead(@(fitParams) evalEllipseFit(zRescaled,fitParams,fixParams,indicatorForParameters),fitParams,optimset('MaxFunEvals',MaxFunEvalOptSet,'MaxIter',MaxNelderMeadIterationsSet,'TolX',xTol,'TolFun',fTol,'Display',optimsetdisplay));%,optimset('Display','iter'));
        counterIterationsForEachScale(scaleIndex)=counterIterationsForEachScale(scaleIndex)+OUTPUT.iterations;
        counterFunctionEvaluationsForEachScale(scaleIndex)=counterFunctionEvaluationsForEachScale(scaleIndex)+OUTPUT.funcCount;
        EllipseParameters(indicatorForParameters)=fitParams;
        
        counterLogLine=counterLogLine+1;
        EvolutionLog(counterLogLine,:)=[scaleIndex downscaleFactor(scaleIndex) counterNelderMeadRestarts FunctionValueCurrentBest counterIterationsForEachScale(scaleIndex) counterFunctionEvaluationsForEachScale(scaleIndex) (now-preprocessingTime1)*60*60*24 EllipseParametersCurrentBest];
        disp(num2str(EvolutionLog(counterLogLine,:)));
    end
    
    optimizationTime1=now;
    
    
    timeOptimizationS(iiii)=(optimizationTime1-preprocessingTime1)*60*60*24;
    disp(['Optimization completed: image #',num2str([iiii timeOptimizationS(iiii)]),' seconds']);
    disp(' '),disp(' ')
    
    %%% [c1,c2,r1,r2,theta,thicknessFactor]
    disp('Optimized ellipse parameters:')
    disp(['r1=',num2str(EllipseParameters(1)),', ',...
        'r2=',num2str(EllipseParameters(2)),', ',...
        'c1=',num2str(EllipseParameters(3)),', ',...
        'c2=',num2str(EllipseParameters(4)),', ',...
        'theta=',num2str(EllipseParameters(5)),', ',...
        'kappa=',num2str(EllipseParameters(6))])
    ParamsS(iiii,:)=EllipseParameters;
    filenameS{iiii}=fnamme;
    EvolsS{iiii}=EvolutionLog;
    CostS(iiii)=FunctionValueCurrentBest;
    
end


%%% fitting of all images completed


%%% compute biometric data and create output files (images etc.) from the estimated parameters
ParamsSE=ParamsS;
rand('seed',sum(clock));
UIDno=num2str(round(1000+(rand*8999)));
for positive_fraction=76.3/100;
    uniqID=['ID_A',num2str(PA),'B',num2str(PB),'__',UIDno,'__',num2str(round(positive_fraction*100)),'_'];
    info=cell(numel(FilenamesCell),1);
    dx=zeros(numel(FilenamesCell),1);
    OFD=zeros(numel(FilenamesCell),1);
    BPD=zeros(numel(FilenamesCell),1);
    BPDoo=zeros(numel(FilenamesCell),1);
    OFDoo=zeros(numel(FilenamesCell),1);
    SKULL_THICKNESS=zeros(numel(FilenamesCell),1);
    SKULL_THICKNESS_FROM_BPD=zeros(numel(FilenamesCell),1);
    CIRCUMFERENCE_EXACT=zeros(numel(FilenamesCell),1);
    CIRCUMFERENCE_APPROX=zeros(numel(FilenamesCell),1);
    HC=zeros(numel(FilenamesCell),1);
    for  iiii=ToDo;
        fnamme=filenameS{iiii};
        disp([fnamme,'     ',num2str(iiii),'/',num2str(numel(FilenamesCell))]);
        try
            z=im2double(rgb2gray(imread(fnamme)));
            dx(iiii) = 1;
        catch err
            z=im2double(dicomread(fnamme,'frames',1));
            info{iiii} = dicominfo(fnamme);
            dx(iiii) =  10*info{iiii}.SequenceOfUltrasoundRegions.Item_1.PhysicalDeltaX;
        end
        z=zPad(z,PadBorderPixel);
        EllipseParameters=ParamsS(iiii,:);
        c1=EllipseParameters(1);
        c2=EllipseParameters(2);
        r1=EllipseParameters(3);
        r2=EllipseParameters(4);
        theta=EllipseParameters(5);
        thicknessFactor=EllipseParameters(6);
        [size_z_1, size_z_2]=size(z);
        len1=1;
        len2=size_z_2/size_z_1*len1;
        [ellipseFittedDoG, ellipseBinary, r_max_min_emboss]=generateEllipse(c1,c2,r1,r2,theta,thicknessFactor,size_z_1,size_z_2,len1,len2,positive_fraction);
        if r1>r2
            r_1_2_emboss=r_max_min_emboss;
        else
            r_1_2_emboss=r_max_min_emboss(2:-1:1);
        end
        ParamsSE(iiii,end-1:end)=r_1_2_emboss;
        ellipseBinary=zUnPad(ellipseBinary,PadBorderPixel);
        ellBinaryS=false(size(ellipseBinary,1),size(ellipseBinary,2),numel(FilenamesCell));
        ellipseFittedDoG=zUnPad(ellipseFittedDoG,PadBorderPixel);
        
        ellBinaryS(:,:,iiii)=logical(ellipseBinary);
        
        
        %%% compute anatomical measures
        OFD(iiii)=2*max(r1,r2)*dx(iiii)*size_z_1;
        BPD(iiii)=2*min(r1,r2)*dx(iiii)*size_z_1;
        OFDoo(iiii)=2*r_max_min_emboss(1)*dx(iiii)*size_z_1;
        BPDoo(iiii)=2*r_max_min_emboss(2)*dx(iiii)*size_z_1;
        SKULL_THICKNESS(iiii)=2*(r_1_2_emboss(1)-r1)*dx(iiii)*size_z_1;
        SKULL_THICKNESS_FROM_BPD(iiii)=1.31-0.119*BPD(iiii)+0.00472*BPD(iiii)^2-0.00003627*BPD(iiii)^3;
        CIRCUMFERENCE_EXACT(iiii)=circumS(iiii)*dx(iiii)*size_z_1;
        CIRCUMFERENCE_APPROX(iiii)=pi*(BPD(iiii)+OFD(iiii))/2;
        HC(iiii)=pi*(BPDoo(iiii)+OFDoo(iiii))/2;
        
        if writeImagePositiveEllipse
            imwrite(max(0,ellipseFittedDoG/max(ellipseFittedDoG(:))),[uniqID,'_ellipse_fit__',num2str(10000+iiii),'__',fnamme(1:end-4),'.tif']);
        end
        
        if writeImageBinaryEllipse
            imwrite(ellipseBinary,[uniqID,'_ellipse_binary__',num2str(10000+iiii),'__',fnamme(1:end-4),'.tif']);
        end
        
        yellowBrightness=0.5; % set to 1 to have max brightness
        yellowEllipseThickness=0.001;   % 0.001 is a fairly decent number
        [yellowEllipseMiddle]=generateEllipse(c1,c2, r1                   , r2                   ,theta,yellowEllipseThickness,size_z_1,size_z_2,len1,len2,positive_fraction);
        [yellowEllipseOuter]=generateEllipse(c1,c2, r_1_2_emboss(1)      , r_1_2_emboss(2)      ,theta,yellowEllipseThickness,size_z_1,size_z_2,len1,len2,positive_fraction);
        [yellowEllipseInner]=generateEllipse(c1,c2, 2*r1-r_1_2_emboss(1) , 2*r2-r_1_2_emboss(2) ,theta,yellowEllipseThickness,size_z_1,size_z_2,len1,len2,positive_fraction);
        yellowEllipseMiddle=max(0,yellowEllipseMiddle/max(yellowEllipseMiddle(:)))*yellowBrightness;
        yellowEllipseOuter=max(0,yellowEllipseOuter/max(yellowEllipseOuter(:)))*yellowBrightness;
        yellowEllipseInner=max(0,yellowEllipseInner/max(yellowEllipseInner(:)))*yellowBrightness;
        yellowEllipsesRGBimage=repmat(z,[1 1 3]);
        yellowEllipsesRGBimage(:,:,1)=yellowEllipsesRGBimage(:,:,1).*(1-yellowEllipseMiddle)+yellowEllipseMiddle;  yellowEllipsesRGBimage(:,:,2)=yellowEllipsesRGBimage(:,:,2).*(1-yellowEllipseMiddle)+yellowEllipseMiddle;
        yellowEllipsesRGBimage(:,:,1)=yellowEllipsesRGBimage(:,:,1).*(1-yellowEllipseOuter)+yellowEllipseOuter;  yellowEllipsesRGBimage(:,:,2)=yellowEllipsesRGBimage(:,:,2).*(1-yellowEllipseOuter)+yellowEllipseOuter;
        yellowEllipsesRGBimage(:,:,1)=yellowEllipsesRGBimage(:,:,1).*(1-yellowEllipseInner)+yellowEllipseInner;  yellowEllipsesRGBimage(:,:,2)=yellowEllipsesRGBimage(:,:,2).*(1-yellowEllipseInner)+yellowEllipseInner;
        
        figure(1),
        imshow(yellowEllipsesRGBimage,[],'initialmag','fit')
        
        if writeImageYellowEllipses
            imwrite(zUnPad(yellowEllipsesRGBimage,PadBorderPixel),[uniqID,'_ellipse_fitYel__',num2str(10000+iiii),'__',fnamme(1:end-4),'.tif']);
        end
    end
end
